#!/usr/bin/env python
"""ENTRYTYPE constants for bib-dedupe"""
# pylint: disable=too-few-public-methods


ARTICLE = "article"
INPROCEEDINGS = "inproceedings"
BOOK = "book"
INBOOK = "inbook"
PROCEEDINGS = "proceedings"

INCOLLECTION = "incollection"
PHDTHESIS = "phdthesis"
THESIS = "thesis"
MASTERSTHESIS = "masterthesis"
BACHELORTHESIS = "bachelorthesis"
TECHREPORT = "techreport"
UNPUBLISHED = "unpublished"
MISC = "misc"
SOFTWARE = "software"
ONLINE = "online"
CONFERENCE = "conference"
